package main

import (
	"context"
	"cryptic-net/admin"
	"cryptic-net/bootstrap"
	"cryptic-net/daemon"
	"cryptic-net/garage"
	"cryptic-net/nebula"
	"crypto/rand"
	"encoding/hex"
	"errors"
	"fmt"
	"net"
	"os"
	"strings"

	"code.betamike.com/cryptic-io/pmux/pmuxlib"
)

func randStr(l int) string {
	b := make([]byte, l)
	if _, err := rand.Read(b); err != nil {
		panic(err)
	}
	return hex.EncodeToString(b)
}

func readAdmin(path string) (admin.Admin, error) {

	if path == "-" {

		adm, err := admin.FromReader(os.Stdin)
		if err != nil {
			return admin.Admin{}, fmt.Errorf("parsing admin.yml from stdin: %w", err)
		}

		return adm, nil
	}

	f, err := os.Open(path)
	if err != nil {
		return admin.Admin{}, fmt.Errorf("opening file: %w", err)
	}
	defer f.Close()

	return admin.FromReader(f)
}

var subCmdAdminCreateNetwork = subCmd{
	name:  "create-network",
	descr: "Creates a new cryptic network, outputting the resulting admin.yml to stdout",
	do: func(subCmdCtx subCmdCtx) error {

		flags := subCmdCtx.flagSet(false)

		daemonConfigPath := flags.StringP(
			"config-path", "c", "",
			"Optional path to a daemon.yml file to load configuration from.",
		)

		dumpConfig := flags.Bool(
			"dump-config", false,
			"Write the default configuration file to stdout and exit.",
		)

		name := flags.StringP(
			"name", "n", "",
			"Human-readable name to identify the network as.",
		)

		domain := flags.StringP(
			"domain", "d", "",
			"Domain name that should be used as the root domain in the network.",
		)

		ipNetStr := flags.StringP(
			"ip-net", "i", "",
			`IP+prefix (e.g. "10.10.0.1/16") which denotes the IP of this host, which will be the first host in the network, and the range of IPs which other hosts in the network can be assigned`,
		)

		hostName := flags.StringP(
			"hostname", "h", "",
			"Name of this host, which will be the first host in the network",
		)

		if err := flags.Parse(subCmdCtx.args); err != nil {
			return fmt.Errorf("parsing flags: %w", err)
		}

		if *dumpConfig {
			return daemon.CopyDefaultConfig(os.Stdout, envAppDirPath)
		}

		if *name == "" || *domain == "" || *ipNetStr == "" || *hostName == "" {
			return errors.New("--name, --domain, --ip-net, and --hostname are required")
		}

		*domain = strings.TrimRight(strings.TrimLeft(*domain, "."), ".")

		ip, subnet, err := net.ParseCIDR(*ipNetStr)
		if err != nil {
			return fmt.Errorf("parsing %q as a CIDR: %w", *ipNetStr, err)
		}

		if err := validateHostName(*hostName); err != nil {
			return fmt.Errorf("invalid hostname %q: %w", *hostName, err)
		}

		runtimeDirCleanup, err := setupAndLockRuntimeDir()
		if err != nil {
			return fmt.Errorf("setting up runtime directory: %w", err)
		}
		defer runtimeDirCleanup()

		daemonConfig, err := daemon.LoadConfig(envAppDirPath, *daemonConfigPath)
		if err != nil {
			return fmt.Errorf("loading daemon config: %w", err)
		}

		if len(daemonConfig.Storage.Allocations) < 3 {
			return fmt.Errorf("daemon config with at least 3 allocations was not provided")
		}

		nebulaCACreds, err := nebula.NewCACredentials(*domain, subnet)
		if err != nil {
			return fmt.Errorf("creating nebula CA cert: %w", err)
		}

		nebulaHostCreds, err := nebula.NewHostCredentials(nebulaCACreds, *hostName, ip)
		if err != nil {
			return fmt.Errorf("creating nebula cert for host: %w", err)
		}

		nebulaHostSignedPublicCreds, err := bootstrap.NewNebulaHostSignedPublicCredentials(
			nebulaCACreds,
			nebulaHostCreds.Public,
		)

		if err != nil {
			return fmt.Errorf("creating signed public credentials for host: %w", err)
		}

		adminCreationParams := admin.CreationParams{
			ID:     randStr(32),
			Name:   *name,
			Domain: *domain,
		}

		hostBootstrap := bootstrap.Bootstrap{
			AdminCreationParams: adminCreationParams,
			Hosts: map[string]bootstrap.Host{
				*hostName: bootstrap.Host{
					Name: *hostName,
					Nebula: bootstrap.NebulaHost{
						SignedPublicCredentials: nebulaHostSignedPublicCreds,
					},
				},
			},
			HostName: *hostName,
		}

		hostBootstrap.Nebula.CAPublicCredentials = nebulaCACreds.Public
		hostBootstrap.Nebula.HostCredentials = nebulaHostCreds
		hostBootstrap.Nebula.SignedPublicCredentials = nebulaHostSignedPublicCreds

		hostBootstrap.Garage.RPCSecret = randStr(32)
		hostBootstrap.Garage.AdminToken = randStr(32)
		hostBootstrap.Garage.GlobalBucketS3APICredentials = garage.NewS3APICredentials()

		if hostBootstrap, err = mergeDaemonConfigIntoBootstrap(hostBootstrap, daemonConfig); err != nil {
			return fmt.Errorf("merging daemon config into bootstrap data: %w", err)
		}

		nebulaPmuxProcConfig, err := nebulaPmuxProcConfig(hostBootstrap, daemonConfig)
		if err != nil {
			return fmt.Errorf("generating nebula config: %w", err)
		}

		garagePmuxProcConfigs, err := garagePmuxProcConfigs(hostBootstrap, daemonConfig)
		if err != nil {
			return fmt.Errorf("generating garage configs: %w", err)
		}

		pmuxConfig := pmuxlib.Config{
			Processes: append(
				[]pmuxlib.ProcessConfig{
					nebulaPmuxProcConfig,
				},
				garagePmuxProcConfigs...,
			),
		}

		ctx, cancel := context.WithCancel(subCmdCtx.ctx)
		pmuxDoneCh := make(chan struct{})

		fmt.Fprintln(os.Stderr, "starting child processes")
		go func() {
			// NOTE both stdout and stderr are sent to stderr, so that the user
			// can pipe the resulting admin.yml to stdout.
			pmuxlib.Run(ctx, os.Stderr, os.Stderr, pmuxConfig)
			close(pmuxDoneCh)
		}()

		defer func() {
			cancel()
			fmt.Fprintln(os.Stderr, "waiting for child processes to exit")
			<-pmuxDoneCh
		}()

		fmt.Fprintln(os.Stderr, "waiting for garage instances to come online")
		if err := waitForGarageAndNebula(ctx, hostBootstrap, daemonConfig); err != nil {
			return fmt.Errorf("waiting for garage to start up: %w", err)
		}

		fmt.Fprintln(os.Stderr, "applying initial garage layout")
		if err := garageApplyLayout(ctx, hostBootstrap, daemonConfig); err != nil {
			return fmt.Errorf("applying initial garage layout: %w", err)
		}

		fmt.Fprintln(os.Stderr, "initializing garage shared global bucket")
		err = garageInitializeGlobalBucket(ctx, hostBootstrap, daemonConfig)

		if cErr := (garage.AdminClientError{}); errors.As(err, &cErr) && cErr.StatusCode == 409 {
			return fmt.Errorf("shared global bucket has already been created, are the storage allocations from a previously initialized cryptic-net being used?")

		} else if err != nil {
			return fmt.Errorf("initializing garage shared global bucket: %w", err)
		}

		fmt.Fprintln(os.Stderr, "cluster initialized successfully, writing admin.yml to stdout")

		adm := admin.Admin{
			CreationParams: adminCreationParams,
		}
		adm.Nebula.CACredentials = nebulaCACreds
		adm.Garage.RPCSecret = hostBootstrap.Garage.RPCSecret
		adm.Garage.GlobalBucketS3APICredentials = hostBootstrap.Garage.GlobalBucketS3APICredentials

		if err := adm.WriteTo(os.Stdout); err != nil {
			return fmt.Errorf("writing admin.yml to stdout")
		}

		return nil
	},
}

var subCmdAdminCreateBootstrap = subCmd{
	name:      "create-bootstrap",
	descr:     "Creates a new bootstrap.yml file for a particular host and writes it to stdout",
	checkLock: true,
	do: func(subCmdCtx subCmdCtx) error {

		flags := subCmdCtx.flagSet(false)

		hostName := flags.StringP(
			"hostname", "h", "",
			"Name of the host to generate bootstrap.yml for",
		)

		ipStr := flags.StringP(
			"ip", "i", "",
			"IP of the new host",
		)

		adminPath := flags.StringP(
			"admin-path", "a", "",
			`Path to admin.yml file. If the given path is "-" then stdin is used.`,
		)

		if err := flags.Parse(subCmdCtx.args); err != nil {
			return fmt.Errorf("parsing flags: %w", err)
		}

		if *hostName == "" || *ipStr == "" || *adminPath == "" {
			return errors.New("--hostname, --ip, and --admin-path are required")
		}

		if err := validateHostName(*hostName); err != nil {
			return fmt.Errorf("invalid hostname %q: %w", *hostName, err)
		}

		ip := net.ParseIP(*ipStr)

		if ip == nil {
			return fmt.Errorf("invalid ip %q", *ipStr)
		}

		adm, err := readAdmin(*adminPath)
		if err != nil {
			return fmt.Errorf("reading admin.yml with --admin-path of %q: %w", *adminPath, err)
		}

		hostBootstrap, err := loadHostBootstrap()
		if err != nil {
			return fmt.Errorf("loading host bootstrap: %w", err)
		}

		nebulaHostCreds, err := nebula.NewHostCredentials(adm.Nebula.CACredentials, *hostName, ip)
		if err != nil {
			return fmt.Errorf("creating new nebula host key/cert: %w", err)
		}

		nebulaHostSignedPublicCreds, err := bootstrap.NewNebulaHostSignedPublicCredentials(
			adm.Nebula.CACredentials,
			nebulaHostCreds.Public,
		)

		if err != nil {
			return fmt.Errorf("creating signed public credentials for host: %w", err)
		}

		newHostBootstrap := bootstrap.Bootstrap{
			AdminCreationParams: adm.CreationParams,

			Hosts:    hostBootstrap.Hosts,
			HostName: *hostName,
		}

		newHostBootstrap.Nebula.CAPublicCredentials = adm.Nebula.CACredentials.Public
		newHostBootstrap.Nebula.HostCredentials = nebulaHostCreds
		newHostBootstrap.Nebula.SignedPublicCredentials = nebulaHostSignedPublicCreds

		newHostBootstrap.Garage.RPCSecret = adm.Garage.RPCSecret
		newHostBootstrap.Garage.AdminToken = randStr(32)
		newHostBootstrap.Garage.GlobalBucketS3APICredentials = adm.Garage.GlobalBucketS3APICredentials

		return newHostBootstrap.WriteTo(os.Stdout)
	},
}

var subCmdAdmin = subCmd{
	name:  "admin",
	descr: "Sub-commands which only admins can run",
	do: func(subCmdCtx subCmdCtx) error {
		return subCmdCtx.doSubCmd(
			subCmdAdminCreateNetwork,
			subCmdAdminCreateBootstrap,
		)
	},
}
