package main

import (
	"errors"
	"fmt"
	"io/fs"
	"os"
	"path/filepath"

	"github.com/shirou/gopsutil/process"
)

var errDaemonNotRunning = errors.New("no cryptic-net daemon process running")

func lockFilePath() string {
	return filepath.Join(envRuntimeDirPath, "lock")
}

func writeLock() error {

	lockFilePath := lockFilePath()

	lockFile, err := os.OpenFile(
		lockFilePath, os.O_WRONLY|os.O_CREATE|os.O_EXCL, 0400,
	)

	if errors.Is(err, os.ErrExist) {
		return fmt.Errorf(
			"lock file %q already exists, if the cryptic-net daemon is not already running you can safely delete this file",
			lockFilePath,
		)

	} else if err != nil {
		return fmt.Errorf("opening lockfile %q: %w", lockFilePath, err)
	}

	defer lockFile.Close()

	if _, err := fmt.Fprintf(lockFile, "%d\n", os.Getpid()); err != nil {
		return fmt.Errorf("writing pid to %q: %w", lockFilePath, err)
	}

	return nil
}

// returns a cleanup function which will clean up the created runtime directory.
func setupAndLockRuntimeDir() (func(), error) {

	fmt.Fprintf(os.Stderr, "will use runtime directory %q for temporary state\n", envRuntimeDirPath)

	if err := os.MkdirAll(envRuntimeDirPath, 0700); err != nil {
		return nil, fmt.Errorf("creating directory %q: %w", envRuntimeDirPath, err)

	} else if err := writeLock(); err != nil {
		return nil, err
	}

	return func() {
		fmt.Fprintf(os.Stderr, "cleaning up runtime directory %q\n", envRuntimeDirPath)
		if err := os.RemoveAll(envRuntimeDirPath); err != nil {
			fmt.Fprintf(os.Stderr, "error removing temporary directory %q: %v", envRuntimeDirPath, err)
		}
	}, nil
}

// checks that the lock file exists and that the process which created it also
// still exists.
func assertLock() error {

	lockFilePath := lockFilePath()

	lockFile, err := os.Open(lockFilePath)

	if errors.Is(err, fs.ErrNotExist) {
		return errDaemonNotRunning

	} else if err != nil {
		return fmt.Errorf("checking lock file %q: %w", lockFilePath, err)
	}

	defer lockFile.Close()

	var pid int32

	if _, err := fmt.Fscan(lockFile, &pid); err != nil {
		return fmt.Errorf("scanning pid from lock file %q: %w", lockFilePath, err)
	}

	procExists, err := process.PidExists(pid)

	if err != nil {
		return fmt.Errorf("checking if process %d exists: %w", pid, err)

	} else if !procExists {
		return errDaemonNotRunning
	}

	return nil
}
